/**
 * Simple MPI matrixes composing.
 *
 * Copyright, Kirill Gavrilov ' 2008
 */

#include <mpi.h>        // Header for MPI (cluster computings)
#include <iostream>     // Header just for base output
#include <fstream>      // Header for file IO

#include "timer.h"      // Header with simple timer
#include "matfile.h"    // Header for Matrix-FILE manipulations

#define WORKTAG 1       // tag to SLAVE: work it!
#define DIETAG 2        // tag to SLAVE: die!

// Local functions
static void master(void);
static void slave(void);

/**
 * Main function, executed on all mashines.
 * Initialize MPI, then divide
 * to master or slave instructions.
 */
int main(int argc, char **argv) {

    // Initialize MPI
    MPI_Init(&argc, &argv);

    // Find out my identity in the default communicator
    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    if (myrank == 0) {
        master();
    } else {
        slave();
    }

    // Shut down MPI
    MPI_Finalize();
    return 0;
}

/**
 * Function runned on master mashine
 * (have a dispetcher function)
 */
static void master(void) {

    std::cout << "Welcome to matrix composing MPI lab by Kirill Gavrilov ' 2008 year!\n";

    int hostsNumber;         // total hosts number
    int hostRank;            // rank variable for cycles
    MPI_Status status;       // status variable for messages
    SimpleTimer timer;       // simple timer

    std::ofstream outFile;   // output file descriptor
    outFile.open("out.dat");
    if (outFile.fail()) {
        std::cout << "ERROR: can't open file for writing results!\n";
        return;
    }

    // Find out how many processes there are in the default communicator
    MPI_Comm_size(MPI_COMM_WORLD, &hostsNumber);
    std::cout << "Master runned. Slaves number: " <<  hostsNumber - 1 << std::endl;

    MatFileReader inMatrix1File("in1.dat"); // open file with first input matrix
    MatFileReader inMatrix2File("in2.dat"); // open file with second input matrix

    struct {
        int height;
        int width;
        int size;
    } matSize, matSizeToSlave, matSizeMaster;      // structures for matrix sizes

    int index = 0;                      // cycle variable
    int recSize = 0;                    // variable for recieved size

    matSize.height = inMatrix1File.getHeight();    // input matrix height
    matSize.width  = inMatrix1File.getWidth();     // input matrix width
    matSize.size = matSize.height * matSize.width; // input matrix size

    std::cout << "Matrixes in input files: " << matSize.height << "x" << matSize.width << std::endl;

    matSizeToSlave.width = matSize.width;                         // set width for piece as input matrix
    matSizeToSlave.height = matSize.height / hostsNumber;         // divide input matrix vertically
    matSizeToSlave.size = matSize.width * matSizeToSlave.height;  // compute size for piece

    // Init memory for sended pieces from first/second input matrix and one piece to recive
    double *matrix1Part = new double[matSizeToSlave.size];
    double *matrix2Part = new double[matSizeToSlave.size];
    double *resultFromSlave = new double[matSizeToSlave.size];

    // Master piece may be bigger then slaves to produce correct data on data/processes combinations
    matSizeMaster.height = matSize.height - matSizeToSlave.height * (hostsNumber - 1);
    matSizeMaster.width = matSize.width;
    matSizeMaster.size = matSizeMaster.width * matSizeMaster.height;

    double *resultMaster = new double[matSizeMaster.size];

    // In loop sent instructions per slave
    for (hostRank = 1; hostRank < hostsNumber; ++hostRank) {

        // Read pieces from input files
        inMatrix1File.getLines((hostRank - 1) * matSizeToSlave.height, (hostRank - 1) * matSizeToSlave.height + matSizeToSlave.height - 1, matrix1Part);
        inMatrix2File.getLines((hostRank - 1) * matSizeToSlave.height, (hostRank - 1) * matSizeToSlave.height + matSizeToSlave.height - 1, matrix2Part);

        // Firstly send to slave matrix size for memory initialization
        MPI_Send(&matSizeToSlave.size,  // message buffer
                 1,                 // one data item
                 MPI_INT,           // data item is an integer
                 hostRank,          // destination process rank
                 WORKTAG,           // WORK tag
                 MPI_COMM_WORLD);   // default communicator

        // Now send part of first matrix
        MPI_Send(matrix1Part, matSizeToSlave.size,  MPI_DOUBLE, hostRank, WORKTAG, MPI_COMM_WORLD);
        // ...and part of second matrix
        MPI_Send(matrix2Part, matSizeToSlave.size,  MPI_DOUBLE, hostRank, WORKTAG, MPI_COMM_WORLD);
    }

    // Compute on master too (LAST piece in data file!)
    inMatrix1File.getLines((hostsNumber - 1) * matSizeToSlave.height, (hostsNumber - 1) * matSizeToSlave.height + matSizeMaster.height - 1, matrix1Part);
    inMatrix2File.getLines((hostsNumber - 1) * matSizeToSlave.height, (hostsNumber - 1) * matSizeToSlave.height + matSizeMaster.height - 1, matrix2Part);
    for (index = 0; index < matSizeMaster.size; index++) {
        resultMaster[index] = matrix1Part[index] + matrix2Part[index];
    }

    // Get results from the slaves and save them to output file
    for (hostRank = 1; hostRank < hostsNumber; ++hostRank) {
        MPI_Recv(resultFromSlave, matSizeToSlave.size, MPI_DOUBLE, hostRank,
                 MPI_ANY_TAG, MPI_COMM_WORLD, &status);

        //MPI_Get_count(&status, MPI_DOUBLE, &recSize);
        //std::cout << "MPI_Get_count=" << recSize << std::endl;
        for (int line = 0; line < matSizeToSlave.height; line++) {
            for (int index = line * matSizeToSlave.width; index < (line + 1) * matSizeToSlave.width; index++) {
                outFile << resultFromSlave[index] << ' ';
            }
            outFile << '\n';
        }
    }

    // Save last piece computed by master
    for (int line = 0; line < matSizeMaster.height; line++) {
        for (int index = line * matSizeMaster.width; index < (line + 1) * matSizeMaster.width; index++) {
            outFile << resultMaster[index] << ' ';
        }
        outFile << '\n';
    }

    unsigned int computTime = timer.getInterval();
    std::cout << "Computings finished. Send kill messages to slaves.\n";

    // Tell all the slaves to exit by sending an empty message with the DIETAG
    for (hostRank = 1; hostRank < hostsNumber; ++hostRank) {
        MPI_Send(0, 0, MPI_INT, hostRank, DIETAG, MPI_COMM_WORLD);
    }

    std::cout << "Program exit\n";
    return;
}

/**
 * Function runned on slave mashines
 * (wait computing rules fron master/dispetcher
 * then send back results)
 */
static void slave(void) {

    double *matrix1Part;    // pointer for part of first matrix
    double *matrix2Part;    // pointer for part of second matrix

    int size = 0;           // matrix dimension
    int index = 0;          // variable for cycle
    MPI_Status status;      // status for messages

    while (true) {

        // Receive matrix size
        MPI_Recv(           &size, // message buffer
                                1, // maxcount
                          MPI_INT, // message DataType
                                0, // source (from whom wait message - from master)
                      MPI_ANY_TAG, // listen only for those tag (any)
                   MPI_COMM_WORLD, // common
                          &status);

        // Check the tag of the received message
        if (status.MPI_TAG == DIETAG) {
            return;
        }

        // Now initialize memory
        matrix1Part = new double[size];
        matrix2Part = new double[size];

        // Receive part of first matrix
        MPI_Recv(matrix1Part, size, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
        // ...and part of second matrix
        MPI_Recv(matrix2Part, size, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

        // Computing...
        for (index = 0; index < size; index++) {
            matrix1Part[index] += matrix2Part[index];
        }

        // Send computing results
	    MPI_Send(matrix1Part, size, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);

        // Clear memory
        delete [] matrix1Part;
    }
}
